import sphobjinv as soi

from .data import Database, Link, Source, Tag


class Collection:
    """A class used for bookeeping."""

    def __init__(self, db, imp_name):
        self.db = db
        self.imp_name = imp_name
        self.source = Source()
        self.links = []
        self.tag_cache = {}
        self.tags = {}

    @property
    def name(self):
        return self.source.name

    @name.setter
    def name(self, value):
        self.source.name = value

    @property
    def prefix(self):
        """The common prefix to every link in the collection"""
        return self.source.prefix

    @prefix.setter
    def prefix(self, value):
        self.source.prefix = value

    def _get_or_create_tag(self, name):
        """Try and find an existing tag with the name, else create it."""

        if name in self.tags:
            return self.tags[name]

        if name in self.tag_cache:
            return self.tag_cache[name]

        existing = Tag.get(self.db, name=name)

        if existing is not None:
            self.tag_cache[name] = existing
            return existing

        tag = Tag(name=name)
        self.tags[name] = tag
        return tag

    def add_link(self, name=None, url=None, tags=None):
        """Handles the detail of making links."""

        link = Link(name=name, url=url)

        if tags is None:
            tags = []

        tags.append(self.imp_name)

        for name in tags:
            tag = self._get_or_create_tag(name)
            link.tags.append(tag)

        self.links.append(link)


def define_importer(import_):
    """Function that handles the details of importing a list of links.

    The idea is that this function calls :code:`f` with the reference given on the
    command line to get the list of links to import. It then handles the details of
    updating the database with these links.
    """

    # This outer function needs to handle the args given on the command line.
    def link_importer(filepath, uri):

        db = Database(filepath, create=True)
        collection = Collection(db, import_.__name__)

        import_(uri, collection)

        source = collection.source
        source.uri = f"{import_.__name__}://{uri}"

        for l in collection.links:
            source.links.append(l)

        Source.add(db, items=[source], commit=False)
        Link.add(db, items=collection.links, commit=False)

        if len(collection.tags) > 0:
            Tag.add(db, items=list(collection.tags.values()), commit=False)

        db.commit()
        db.close()

    return link_importer


def importer(f=None):
    """Decorator for defining importers."""

    if f is None:
        return define_importer

    return define_importer(f)


@importer
def sphinx(uri: str, collection: Collection):
    """Import links from a sphinx documentation site."""

    if uri.endswith("/"):
        uri = f"{uri}objects.inv"

    if not uri.endswith("objects.inv"):
        uri = f"{uri}/objects.inv"

    print(f"Trying index url  : {uri}", end="\r")
    inv = soi.Inventory(url=uri)
    print(f"Found object index: {uri} with {inv.count} entries")

    collection.name = f"{inv.project} v{inv.version} Documentation"
    collection.prefix = uri.replace("objects.inv", "")

    for item in inv.objects:

        name = item.dispname_expanded
        url = item.uri_expanded
        tags = [t for t in [item.domain, item.role] if t is not None]

        collection.add_link(name=name, url=url, tags=tags)



import argparse
import collections
import inspect
import logging
import pathlib
import shutil
import sys

import appdirs
import pkg_resources

from llyfrau._version import __version__
from llyfrau.data import Database, Link, Source, Tag

from .tui import LinkTable

_LogConfig = collections.namedtuple("LogConfig", "level,fmt")
_LOG_LEVELS = [
    _LogConfig(level=logging.INFO, fmt="%(message)s"),
    _LogConfig(level=logging.DEBUG, fmt="[%(levelname)s]: %(message)s"),
    _LogConfig(level=logging.DEBUG, fmt="[%(levelname)s][%(name)s]: %(message)s"),
]


def _setup_logging(verbose: int, quiet: bool) -> None:
    """Setup the logging system according to the given args."""

    if quiet:
        return

    verbose = 0 if verbose < 0 else verbose

    try:
        conf = _LOG_LEVELS[verbose]
        others = False
    except IndexError:
        conf = _LOG_LEVELS[-1]
        others = True

    logger = logging.getLogger("llyfrau")
    logger.setLevel(conf.level)

    console = logging.StreamHandler()
    console.setFormatter(logging.Formatter(conf.fmt))
    logger.addHandler(console)

    if others:
        sql_logger = logging.getLogger("sqlalchemy")
        sql_logger.setLevel(logging.INFO)
        sql_logger.addHandler(console)


def add_link(filepath, url, name, tags):
    db = Database(filepath, create=True)

    if tags is None:
        Link.add(db, name=name, url=url)
        return 0

    link = Link(name=name, url=url)
    new_tags = []

    for t in tags:
        existing = Tag.get(db, name=t)

        if existing is not None:
            link.tags.append(existing)
            continue

        tag = Tag(name=t)
        link.tags.append(tag)
        new_tags.append(tag)

    Link.add(db, items=[link], commit=False)

    if len(new_tags) > 0:
        Tag.add(db, items=new_tags, commit=False)

    db.commit()


def find_sources(filepath):

    path = pathlib.Path(filepath)

    if not path.exists():
        print(f"Unable to find links database: {filepath}", file=sys.stderr)
        return -1

    ids = ["ID"]
    names = ["Name"]
    uris = ["URI"]
    prefixes = ["Prefix"]

    db = Database(filepath)

    for source in Source.search(db):
        ids.append(source.id)
        names.append(source.name)
        uris.append(source.uri)
        prefixes.append(source.prefix)

    print(format_table([ids, names, uris, prefixes]))


def open_link_ui(filepath):

    table_ui = LinkTable(filepath)
    table_ui.run()


def call_command(cmd, args):

    if args.filepath is None:
        base = appdirs.user_data_dir(appname="llyfr", appauthor=False)
        args.filepath = str(pathlib.Path(base, "links.db"))

    params = inspect.signature(cmd).parameters
    cmd_args = {name: getattr(args, name) for name in params}
    return cmd(**cmd_args)


def format_cell(text, width, placeholder=None):

    if placeholder is None:
        placeholder = "..."

    if len(text) <= width:
        return text.ljust(width)

    idx = width - len(placeholder)
    return text[:idx] + placeholder


def format_column(col):
    return ["" if c is None else str(c) for c in col]


def format_table(columns):

    term_width, _ = shutil.get_terminal_size()
    maxwidth = (term_width - len(columns) + 1) // len(columns)

    columns = [format_column(col) for col in columns]
    sizes = [max(len(c) for c in col) for col in columns]
    widths = [min(maxwidth, size) for size in sizes]

    fmt = " ".join(["{:<" + str(n) + "}" for n in widths])
    rows = []

    for i in range(len(columns[0])):
        row = [
            format_cell(col[i], width=widths[idx]) for idx, col in enumerate(columns)
        ]
        rows.append(fmt.format(*row))

    return "\n".join(rows)


def _load_importers(parent):
    """Load importers and attach them to the cli interface."""

    for imp in pkg_resources.iter_entry_points("llyfrau.importers"):
        cmd = parent.add_parser(imp.name, help=f"{imp.name} importer")
        cmd.add_argument("uri", help="where to import links from")

        impl = imp.load()
        cmd.set_defaults(run=impl)


cli = argparse.ArgumentParser()
cli.add_argument(
    "-f", "--filepath", type=str, help="filepath to the links database", default=None,
)
cli.add_argument(
    "-q", "--quiet", help="disable all console output", action="store_true"
)
cli.add_argument(
    "-v",
    "--verbose",
    action="count",
    default=0,
    help="increase output verbosity, repeatable e.g. -v, -vv, -vvv, ...",
)
cli.add_argument("--version", action="store_true", help="show version and exit")

commands = cli.add_subparsers(title="commands")
add = commands.add_parser("add", help="add a link")
add.add_argument("url", help="the link to add")
add.add_argument("-n", "--name", help="name of the link")
add.add_argument("-t", "--tags", nargs="*", help="tags to apply to the link")
add.set_defaults(run=add_link)

import_ = commands.add_parser("import", help="import links from a source")
importers = import_.add_subparsers(title="importers")
_load_importers(importers)

sources = commands.add_parser("sources", help="list all link sources")
sources.set_defaults(run=find_sources)

open_ = commands.add_parser("open", help="open a link")
open_.set_defaults(run=open_link_ui)


def main():

    args = cli.parse_args()

    if args.version:
        print(f"llyfr v{__version__}")
        return 0

    _setup_logging(args.verbose, args.quiet)

    if hasattr(args, "run"):
        return call_command(args.run, args)

    cli.print_help()
